open Printf
open Expression
open ExpNode
open BinaryMatrix
open BinaryMatrixSet
open NodeSet
open EdgeSet
open NodeSet2
open NodeSetSet
open Graph
open Options

	exception NONSENSEGRAPH of string
	exception INVALIDCONTEXT

       type exp = Expression.exp
	type allNodeSet = NodeSet.t 
	type allEdgeSet = EdgeSet.t
	type allNodeList = Graph.node list
	type allEdgeList = Graph.edge list
       type nodeSetSet = NodeSetSet.t	
	type state = int * exp ref list ref       
	type graphMIB = { numof : int; initVal : exp ref list ref; nodeList : allNodeList ref; edgeList : allEdgeList ref; equivNodeSets : nodeSetSet ref; cmG : BinaryMatrix.binaryMatrix}
    	type symbolicGraph = allNodeSet ref * allEdgeSet ref

	let graphInfo = ref { numof = 0; initVal = ref []; nodeList = ref []; edgeList = ref []; equivNodeSets = ref NodeSetSet.empty; cmG = BinaryMatrix.make 0 0 false }

	let tauPrunning ns es ea =
		let moreRounds = ref true
		and seqno = ref (List.length !ea) in
		let prune sid b assign tid deltau eid=
			let (sid1, b1, assign1, alpha1, tid1) = List.nth !ea eid
			in if sid1 = tid then begin
				if ((sid != tid) || (!alpha1 != TAU)) then begin
					deltau := true; 
					let b' = conjunctBoolExp b (substCondition b1 assign)
					and assign' = mergeAssign assign assign1
					in begin
(* 						printf "%d-(%s,%s,%s)->%d\n" sid (Expression.exp2String b') (Graph.assign2String assign') (Graph.action2String alpha1) tid1; *)
						ea := List.append !ea [(sid, b', assign', alpha1, tid1)];
						es := EdgeSet.add !seqno !es;
						seqno := !seqno + 1;
						moreRounds := true
					end
				end
				else
					deltau := false
			end in
		let findTau eid =
			let deltau = ref true
			and (sid, b, assign, alpha, tid) = List.nth !ea eid
			in match !alpha with
				|TAU ->	if (sid != tid) then
						begin
							EdgeSet.iter (prune sid b assign tid deltau) !es;
							if (!deltau) then
								es := EdgeSet.remove eid !es
							else
								alpha := DELTA(0)
						end
						else
							alpha := DELTA(1)
				|_ -> ()
		in while (!moreRounds) do
			moreRounds := false;
			EdgeSet.iter findTau !es;
		done

	let buildCM ns es n ea =
		let cm = BinaryMatrix.make n n false in
		let setEdge eid =
			let (sid, _, _, _, tid) = List.nth !ea eid
			in	if sid >=0 && sid < n && tid >= 0 && tid < n then
					BinaryMatrix.set cm sid tid true
				else
					raise (NONSENSEGRAPH ("buildSizeCM: " ^ (string_of_int sid) ^ " -> " ^ (string_of_int tid) ^ " in " ^ (string_of_int n) ^ " node(s)"))
		in	tauPrunning ns es ea;
			EdgeSet.iter setEdge !es;
			cm
			
	let buildActionCM (ns, es) alpha =
		let n = (!graphInfo).numof 
		and ea = (!graphInfo).edgeList in
		let cm = BinaryMatrix.make n n false in
		let setEdge eid =
			let (sid, _, _, beta, tid) = List.nth !ea eid in
			let res = Graph.is_act_eq alpha beta
			in	if res then begin
					if sid < n && tid < n then
						BinaryMatrix.set cm sid tid true
					else
						raise (NONSENSEGRAPH ("buildActionCM: " ^ (string_of_int sid) ^ " -> " ^ (string_of_int tid) ^ " in " ^ (string_of_int n) ^ " node(s)"))
				end
		in	EdgeSet.iter setEdge !es;
			cm

	let reachabilityCheck (ns, es)=
		let reachAll =
			let nsrc = (!graphInfo).numof
			and na = (!graphInfo).nodeList
			and ea = (!graphInfo).edgeList
			and cmG = (!graphInfo).cmG in
			let rec next prov =
				let prov' = BinaryMatrix.add prov (BinaryMatrix.prod prov cmG)
				in if BinaryMatrix.comp prov' prov then prov'
				else next prov'
			and filterNode prov (nid, _) = BinaryMatrix.get prov 0 nid 
			and filterEdge prov eid  = 
				let (sid, _, _, _, tid) = List.nth !ea eid
				in (BinaryMatrix.get prov 0 sid) && (BinaryMatrix.get prov 0 tid)
			and initV = BinaryMatrix.make 1 nsrc false in
			let _ = BinaryMatrix.set initV 0 0 true 
			and pro = next initV
			in 	ns := NodeSet.filter (filterNode pro) !ns;
				es := EdgeSet.filter (filterEdge pro) !es;
		in if ((NodeSet.cardinal !ns) != 0) then reachAll

	let printGraph (ns, es) =
		let na = (!graphInfo).nodeList
		and ea = (!graphInfo).edgeList in
		let printNode (id, ntype) = 
			let (nid, name, _, _) = List.nth !na id
			in match ntype with
				|INIT -> printf "%s:%s is an initial node.\n" nid name
				|PROGRESS -> printf "%s:%s\n" nid name
				|ACCEPTANCE -> printf "%s:%s is an acceptance node.\n" nid name
		and printEdge eid =
			let (sid, b, assign, alpha, tid) = List.nth !ea eid in
			let (ssid, _, _, _) = List.nth !na sid
			and (ttid, _, _, _) = List.nth !na tid 
			in printf "%s-(%s,%s,%s)->%s\n" ssid (Expression.exp2String b) (Graph.assign2String assign) (Graph.action2String alpha) ttid in 
		let rec printSet printMember nl = match nl with
			[] -> ()
			| n::others -> printMember n; printSet printMember others
		in	printf "A STGA model is ready with %d node(s) and %d edge(s).\n" (NodeSet.cardinal !ns) (EdgeSet.cardinal !es);
			printSet printNode (NodeSet.elements !ns);
			printf ";\n";
			printSet printEdge (EdgeSet.elements !es)

	let rec edgeList2String el na ea =
	match el with
		|[] -> "" 
		|eid::rest ->
			let (sid, _, _, _, tid) = List.nth !ea eid in
			let (nsid, _, _, _) = List.nth !na sid
			and (ntid, _, _, _) = List.nth !na tid 
			in match rest with
				|[] -> (*(string_of_int eid) ^ ":" ^ *)nsid ^ "->" ^ ntid
				|_ -> (*(string_of_int eid) ^ ":" ^ *)nsid ^ "->" ^ ntid ^ "," ^ (edgeList2String rest na ea)

	let getNextState nid eval =
		let na = (!graphInfo).nodeList in
		let (_, _, args, _) = List.nth !na nid
		in 	if (List.length !args) = 0 then 
				(nid, ref[])
			else match !eval with
				|ASSIGN(vl, el) -> (nid, ref (List.map (substExp vl el) !args))
				|_ -> raise INVALIDCONTEXT

	let compareState (st1, arg1) (st2, arg2) =
		let isEquivNodes nid1 nid2 nns =
			let (_, ns) = nns
			in NodeSet2.mem nid1 !ns && NodeSet2.mem nid2 !ns in
		let rec compareValueList vall1 vall2 = match (vall1, vall2) with
			|([], []) -> 0
			|(val1::l1, val2::l2) -> begin
				match (!val1, !val2) with
					|(VALUE(v1), VALUE(v2)) ->
						let res = compare v1 v2
						in if res = 0 then compareValueList l1 l2 else res
					|(BVALUE(v1), BVALUE(v2)) ->
						let res = compare v1 v2
						in if res = 0 then compareValueList l1 l2 else res
					|(VAR(v1), VAR(v2)) ->
						(*let res = compare v1 v2
						in if res = 0 then*) compareValueList l1 l2 (*else res*)
					|(VALUE(v1), VAR(v2)) ->
						compare (string_of_int v1) v2
					|(VAR(v1), VALUE(v2)) ->
						compare v1 (string_of_int v2)
					|(BVALUE(v1), VAR(v2)) ->
						compare (string_of_bool v1) v2
					|(VAR(v1), BVALUE(v2)) ->
						compare v1 (string_of_bool v2)
					|_ -> 1 (* raise (INVALIDSTATE((string_of_int st1)^":"^(expList2String arg1)^" vs. "^(string_of_int st1)^":"^(expList2String arg2))) *)
						(* begin printf "INVALIDSTATE:\n%s\n" ((string_of_int st1)^":"^(expList2String arg1)^" vs.\n"^(string_of_int st2)^":"^(expList2String arg2)); 1 end *)
				end
			|_ -> raise (INVALIDSTATE((string_of_int st1)^":"^(expList2String arg1)^" vs. "^(string_of_int st1)^":"^(expList2String arg2)))
		and res = compare st1 st2
		in	(*printf "%s\n" ((string_of_int st1)^":"^(expList2String arg1)^" vs.\n"^(string_of_int st2)^":"^(expList2String arg2));*)
			if res != 0 then
				let eq = NodeSetSet.exists (isEquivNodes st1 st2) !((!graphInfo).equivNodeSets)
				in begin
					if eq then
						compareValueList !arg1 !arg2
					else
						res
				end
			else
				compareValueList !arg1 !arg2

	let printState (id, vall) =
		let (nid, _, _, _) = List.nth !((!graphInfo).nodeList) id
		in printf "Visited State: %s(%s).\n" nid (expList2String vall)
	
	let logState (id, vall) =
		let (nid, _, _, _) = List.nth !((!graphInfo).nodeList) id
		in fprintf !(Options.logFile) "New State: %s(%s).\n" nid (expList2String vall)

	let getVarNameList vn varNameList =
		varNameList := List.append !varNameList [ref (VAR(vn))]
	
	let getVarValueList vv varValueList =
		varValueList := List.append !varValueList [ref (VALUE(vv))]

